/* PlotBand.java

	Purpose:
		
	Description:
		
	History:
		Fri, Jan 17, 2014 12:02:18 PM, Created by RaymondChao

Copyright (C) 2014 Potix Corporation. All Rights Reserved.

 */
package org.zkoss.chart;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.zkoss.chart.OptionDataEvent.EventType;
import org.zkoss.chart.util.DynamicalAttribute;
import org.zkoss.json.JSONObject;
import org.zkoss.json.JavaScriptValue;
import org.zkoss.util.CollectionsX;
import org.zkoss.zk.ui.event.Event;
import org.zkoss.zk.ui.event.EventListener;
import org.zkoss.zk.ui.event.Events;

/**
 * A colored band stretching across the plot area marking an interval on the
 * axis. In a gauge, a plot band on the Y axis (value axis) will stretch along
 * the perimiter of the gauge.
 * <p>
 * All the options in this class support {@link DynamicalAttribute}.
 * 
 * @author RaymondChao
 * @author jumperchen
 */
public class PlotBand extends Optionable implements ClientEventHandler {
	private enum Attrs implements PlotAttribute, DynamicalAttribute {
		color, from, id, innerRadius, label, outerRadius, to, zIndex
	}

	/**
	 * Default constructor
	 */
	public PlotBand() {
	}

	/**
	 * Constructs with the values, from, to, and color.
	 */
	public PlotBand(Number from, Number to, String color) {
		setFrom(from);
		setTo(to);
		setColor(color);
	}

	/**
	 * Returns the color of the plot band.
	 * <p>
	 * Default: null.
	 */
	public Color getColor() {
		return (Color) getAttr(Attrs.color, null).asValue();
	}

	/**
	 * Sets the color of the plot band.
	 */
	public void setColor(Color color) {
		setAttr(Attrs.color, color);
	}

	/**
	 * Sets the color of the plot band.
	 */
	public void setColor(String color) {
		setColor(new Color(color));
	}

	/**
	 * Sets the color of the plot band.
	 */
	public void setColor(LinearGradient color) {
		setColor(new Color(color));
	}

	/**
	 * Sets the color of the plot band.
	 */
	public void setColor(RadialGradient color) {
		setColor(new Color(color));
	}

	/**
	 * Returns the start position of the plot band in axis units.
	 */
	public Number getFrom() {
		return getAttr(Attrs.from, null).asNumber();
	}

	/**
	 * Sets the start position of the plot band in axis units. 'from the start
	 * position of the plot band in axis units.
	 */
	public void setFrom(Number from) {
		setAttr(Attrs.from, from);
	}

	/**
	 * Returns an id used for identifying the plot band in
	 * {@link Axis#removePlotBand(String)}.
	 */
	public String getId() {
		return getAttr(Attrs.id, null).asString();
	}

	/**
	 * Sets an id used for identifying the plot band in
	 * {@link Axis#removePlotBand(String)}.
	 */
	public void setId(String id) {
		setAttr(Attrs.id, id);
	}

	/**
	 * Returns the inner radius.
	 */
	public Object getInnerRadius() {
		return getAttr(Attrs.innerRadius, null).asValue();
	}

	/**
	 * Sets the inner radius.
	 */
	public void setInnerRadius(String innerRadius) {
		setAttr(Attrs.innerRadius, innerRadius);
	}

	/**
	 * Sets the inner radius.
	 */
	public void setInnerRadius(Number innerRadius) {
		setAttr(Attrs.innerRadius, innerRadius);
	}

	/**
	 * Returns text labels for the plot bands
	 * <p>
	 * Default: an instance of PlotLabel.
	 */
	public PlotLabel getLabel() {
		PlotLabel label = (PlotLabel) this.getAttr(Attrs.label);

		if (label == null) {
			label = new PlotLabel();
			setLabel(label);
		}
		return label;
	}

	/**
	 * Sets text labels for the plot bands
	 */
	public void setLabel(PlotLabel label) {
		setAttr(Attrs.label, label);
	}

	public void setLabel(String text) {
		getLabel().setText(text);
	}

	/**
	 * Returns the outer radius
	 * <p>
	 * Default: null.
	 */
	public Object getOuterRadius() {
		return getAttr(Attrs.outerRadius, null).asValue();
	}

	/**
	 * Sets the outer radius
	 */
	public void setOuterRadius(String outerRadius) {
		setAttr(Attrs.outerRadius, outerRadius);
	}

	/**
	 * Sets the outer radius
	 */
	public void setOuterRadius(Number outerRadius) {
		setAttr(Attrs.outerRadius, outerRadius);
	}

	/**
	 * Returns the end position of the plot band in axis units.
	 */
	public Number getTo() {
		return getAttr(Attrs.to, null).asNumber();
	}

	/**
	 * Sets the end position of the plot band in axis units.
	 */
	public void setTo(Number to) {
		setAttr(Attrs.to, to);
	}

	/**
	 * Returns the z index of the plot band within the chart.
	 * <p>
	 * Default: null
	 */
	public Number getZIndex() {
		return getAttr(Attrs.zIndex, null).asNumber();
	}

	/**
	 * Sets the z index of the plot band within the chart.
	 */
	public void setZIndex(Number zIndex) {
		setAttr(Attrs.zIndex, zIndex);
	}
	
	private Map<String, List<EventListener<? extends Event>>> _events;

	/**
	 * Adds an event listener to this options.
	 * @param evtnm what event to listen (never null), i.e. onClick, onMouseOver, and so on.
	 * @since 1.2.0
	 */
	public void addEventListener(String evtnm,
			EventListener<? extends Event> listener) {
		if (!Events.isValid(evtnm))
			throw new IllegalArgumentException("Invalid event name: "+evtnm);
		if (_events == null)
			_events = new HashMap<String, List<EventListener<? extends Event>>>(
					3);
		
		List<EventListener<? extends Event>> list = _events.get(evtnm);
		if (list == null) {
			list = new LinkedList<EventListener<? extends Event>>();
			list.add(listener);
			_events.put(evtnm, list);
		} else {
			list.add(listener);
		}
		fireEvent(EventType.REGISTER, evtnm, getEventListeners(evtnm), null);
	}
	
	/**
	 * Removes an event listener.
	 * @return whether the listener is removed; false if it was never added.
	 * @since 1.2.0
	 */
	public boolean removeEventListener(String evtnm, EventListener<? extends Event> listener) {
		if (evtnm == null || listener == null)
			throw new IllegalArgumentException("null");
		List<EventListener<? extends Event>> list = _events == null ? null : _events.get(evtnm);
		if (list != null) {
			for (Iterator<EventListener<? extends Event>> it = list.iterator(); it.hasNext();) {
				EventListener<? extends Event> next = it.next();
				if (next.equals(listener))
					it.remove();
				if (list.isEmpty()) {
					_events.remove(evtnm);
					fireEvent(EventType.UNREGISTERED, evtnm, null, null);
				} else {
					// update list
					fireEvent(EventType.REGISTER, evtnm, getEventListeners(evtnm), null);
				}

				return true;
			}
		}
		return false;
	}
	/**
	 * Returns all event names, if any.
	 * @since 1.2.0
	 */
	@SuppressWarnings("unchecked")
	public Set<String> getEventNames() {
		if (_events == null)
			return (Set<String>) Collections.EMPTY_SET;
		return _events.keySet();
	}
	/** Returns an iterative collection of the event listeners for the given event.
	 * <p>Note: it is OK to invoke {@link #addEventListener} or {@link #removeEventListener}
	 * when iterating through the event listeners with the returned collection.
	 * <p>To remove an event listener from the returned iterative collection,
	 * you could invoke {@link Iterable#iterator}'s {@link Iterator#remove}.
	 * @since 1.2.0
	 */
	public Iterable<EventListener<? extends Event>> getEventListeners(String evtnm) {
		final List<EventListener<? extends Event>> lis = _events == null ? null :_events.get(evtnm);
		if (lis != null)
			return new Iterable<EventListener<? extends Event>>() {
				public Iterator<EventListener<? extends Event>> iterator() {
					return CollectionsX.comodifiableIterator(lis);
				}
			};
		return CollectionsX.emptyIterable();
	}
	
	public String toJSONString() {
		if (_events == null || _events.isEmpty()) {
			return JSONObject.toJSONString(options);
		} else {
			HashMap<String, Object> clone = new HashMap<String, Object>(options);
			clone.put("uuid", hashCode());
			JSONObject events = new JSONObject();
			clone.put("events", events);
			for (String nm : getEventNames()) {
				nm = nm.substring(2).toLowerCase();
				events.put(
						nm,
						new JavaScriptValue(
								"function (e) {var p = this.axis.chart.pointer; if (!p || !p.hasDragged) zk.Widget.$(e.target).sendDomEvent(e, this);}"));
			}
			return JSONObject.toJSONString(clone);
		}
	}
}