/** Series.java.

	Purpose:
		
	Description:
		
	History:
		2:06:58 PM Jan 9, 2014, Created by jumperchen

Copyright (C) 2014 Potix Corporation. All Rights Reserved.
 */
package org.zkoss.chart;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.zkoss.chart.OptionDataEvent.EventType;
import org.zkoss.chart.plotOptions.DataLabels;
import org.zkoss.chart.plotOptions.SeriesPlotOptions;
import org.zkoss.chart.util.AnyVal;
import org.zkoss.chart.util.DeferredCall;
import org.zkoss.chart.util.DynamicalAttribute;
import org.zkoss.chart.util.JSFunction;
import org.zkoss.json.JSONObject;
import org.zkoss.lang.Generics;
import org.zkoss.lang.Objects;

/**
 * The Series object is the JavaScript representation of each line, area series,
 * pie etc. The object can be accessed in a number of ways. All series and point
 * event handlers give a reference to the series object. The chart object has a
 * series property that is a collection of all the chart's series. The point
 * objects also have the same reference. Another way to reference the series
 * programmatically is by id.
 * <p>
 * All the options in this class support {@link DynamicalAttribute}.
 * 
 * @author jumperchen
 */
public class Series extends Optionable implements OptionDataListener {
	private enum DAttrs implements PlotAttribute, DynamicalAttribute {
		animation,
		borderWidth,
		color, 
		dataLabels,
		// dataParser, Deprecated
		// dataURL, Deprecated
		colorByPoint, 
		index,
		legendIndex,
		lineWidth,
		marker,
		name,
		overshoot,
		pointInterval,
		pointPadding,
		pointPlacement,
		stack,
		tooltip,
		type,
		xAxis,
		yAxis,
		zIndex,
		id
	}

	private enum Attrs implements PlotAttribute {
		data, visible,

		// internal use
		point
	}

	private SeriesPlotOptions plotOptions;

	public Series() {
	}

	/**
	 * Constructs the series with id and a list of point data.
	 */
	public Series(String id, List<Point> data) {
		setId(id);
		setData(data.toArray(new Point[0]));
	}

	/**
	 * Constructs the series with id and point data.
	 */
	public Series(String id, Point... data) {
		setId(id);
		setData(data);
	}

	/**
	 * Returns an id for the series.
	 * <p>
	 * Default: null.
	 */
	public String getId() {
		return getAttr(DAttrs.id, null).asString();
	}

	/**
	 * Sets an id for the series.
	 */
	public void setId(String id) {
		setAttr(DAttrs.id, id);
	}
	
	/**
	 * Returns whether to enable or disable the initial animation when a series is displayed.
	 * <p>
	 * Default: true.
	 */
	public boolean getAnimation() {
		return getAttr(DAttrs.animation, true).asBoolean();
	}
	
	/**
	 * Sets whether to enable or disable the initial animation when a series is displayed.
	 */
	public void setAnimation(boolean animation) {
		setAttr(DAttrs.animation, animation, true);
	}
	
	/**
	 * Returns the border width of the series.
	 */
	public Number getBorderWidth() {
		return getAttr(DAttrs.borderWidth, null).asNumber();
	}

	/**
	 * Sets the border width of the series.
	 */
	public void setBorderWidth(Number borderWidth) {
		setAttr(DAttrs.borderWidth, borderWidth);
	}

	/**
	 * Sets the plot options to the series.
	 */
	public void setPlotOptions(SeriesPlotOptions plotOptions) {
		SeriesPlotOptions old = this.plotOptions;
		if (!Objects.equals(old, plotOptions)) {
			if (old != null)
				old.removeOptionDataListener(this);

			if (plotOptions != null) {
				plotOptions.addOptionDataListener(this);
				this.plotOptions = plotOptions;
			}
		}
	}

	/**
	 * Returns the plot options to the series. If null, it will create a new
	 * one.
	 */
	public SeriesPlotOptions getPlotOptions() {
		if (plotOptions == null) {
			plotOptions = new SeriesPlotOptions();
			setPlotOptions(plotOptions);
		}
		return plotOptions;
	}

	/**
	 * Add a point at the end of the current points list.
	 * 
	 * @param point
	 */
	public void addPoint(Point point) {
		addPoint(point, true, false, true);
	}

	/**
	 * Add a point to the end of the current points list.
	 * 
	 * @param point
	 *            the point
	 * @param redraw
	 *            whether to redraw the chart
	 * @param shift
	 *            If shift is true, a point is shifted off the start of the
	 *            series as one is appended to the end.
	 * @param animation
	 *            Whether to apply animation, and optionally animation
	 *            configuration
	 */
	public void addPoint(Point point, boolean redraw, boolean shift,
			boolean animation) {
		addPoint(point, redraw, shift, animation ? new Animation()
				: Animation.NONE);
	}

	/**
	 * Add a point at the end of the current points list.
	 * 
	 * @param point
	 *            the point
	 * @param redraw
	 *            whether to redraw the chart
	 * @param shift
	 *            If shift is true, a point is shifted off the start of the
	 *            series as one is appended to the end.
	 * @param animation
	 *            Whether to apply animation, and optionally animation
	 *            configuration
	 */
	public void addPoint(final Point point, final boolean redraw,
			final boolean shift, final Animation animation) {
		List<Point> list = getData();
		if (list == null)
			list = new LinkedList<Point>();
		list.add(point);

		// monitor if data changed.
		point.addOptionDataListener(this);
		point.setSeries(this);
		setAttr(Attrs.data, list);

		fireEvent(EventType.ADDED, Attrs.point.toString(), point,
				new DeferredCall() {
					public void execute(JSFunction func) {
						func.callFunction("addPoint", point, redraw, shift,
								animation);
					}
				});
	}

	/**
	 * Add a number point to the end of the current points list.
	 */
	public void addPoint(Number y) {
		if (y == null)
			addPoint(new Point());
		else
			addPoint(new Point(y));
	}

	/**
	 * Add a number point to the end of the current points list.
	 */
	public void addPoint(double y) {
		addPoint(new Point(y));
	}

	/**
	 * Add a point with x and y numbers to the end of the current points list.
	 */
	public void addPoint(Number x, Number y) {
		addPoint(new Point(x, y));
	}

	/**
	 * Add a point with x and y numbers to the end of the current points list.
	 */
	public void addPoint(double x, double y) {
		addPoint(new Point(x, y));
	}

	/**
	 * Add a point with x, low, and high numbers to the end of the current
	 * points list.
	 */
	public void addPoint(Number x, Number low, Number high) {
		addPoint(new Point(x, low, high));
	}

	/**
	 * Add a point with x, low, and high numbers to the end of the current
	 * points list.
	 */
	public void addPoint(double x, double low, double high) {
		addPoint(new Point(x, low, high));
	}

	/**
	 * Add a point with name and number to the end of the current points list.
	 */
	public void addPoint(String name, Number y) {
		addPoint(new Point(name, y));
	}

	/**
	 * Add a point with name, Y number, and X index to the end of the current points list.
	 * @since 1.1
	 */
	public void addPoint(String name, Number y, Number x) {
		addPoint(new Point(name, y, x));
	}

	/**
	 * Add a point with name and number to the end of the current points list.
	 */
	public void addPoint(String name, double y) {
		addPoint(new Point(name, y));
	}

	/**
	 * Sets a list of point data
	 */
	public <T extends Number> void setData(List<T> data) {
		Point[] points = new Point[data.size()];
		int i = 0;
		for (T t : data) {
			if (t == null)
				points[i++] = new Point();
			else
				points[i++] = new Point(t);
		}
		setData(points);
	}

	/**
	 * Sets an array of point data
	 */
	public void setData(Point... data) {
		for (Point p : data) {
			// monitor if data changed.
			p.addOptionDataListener(this);
			p.setSeries(this);
		}

		// use ArrayList instead of fixed ArrayList
		final List<Point> list = new ArrayList<Point>(Arrays.asList(data));
		setAttr(Attrs.data, list);

		fireEvent(EventType.CHANGED, Attrs.data.toString(), data,
				new DeferredCall() {
					public void execute(JSFunction func) {
						func.callFunction("setData", list);
					}
				});
	}

	/**
	 * Sets an array of number data
	 */
	public void setData(Number... data) {
		setData(Arrays.asList(data));
	}

	/**
	 * Sets an array of number data
	 */
	public void setData(Double... data) {
		setData(Arrays.asList(data));
	}

	/**
	 * Sets an array of number data
	 */
	public void setData(Integer... data) {
		setData(Arrays.asList(data));
	}

	/**
	 * Returns the list of point data
	 */
	public List<Point> getData() {
		return Generics.cast(getAttr(Attrs.data, null).asValue());
	}

	/**
	 * Returns the point in the list from the given index.
	 */
	public Point getPoint(int index) {
		return getData().get(index);
	}

	/**
	 * Sets the index of the series in the chart, affecting the internal index
	 * in the chart.series array, the visible Z index as well as the order in
	 * the legend.
	 */
	public void setIndex(int index) {
		if (index < 0) {
			removeKey(DAttrs.index, true);
		} else {
			setAttr(DAttrs.index, index);
		}
	}

	/**
	 * Returns the index of the series in the chart, affecting the internal
	 * index in the chart.series array, the visible Z index as well as the order
	 * in the legend.
	 * <p>
	 * Default: -1, it means depended on the implementaion.
	 */
	public int getIndex() {
		return getAttr(DAttrs.index, -1).asInt();
	}

	/**
	 * Sets the option of data labels.
	 * 
	 * @param dataLabels
	 * @see SeriesPlotOptions#setDataLabels(DataLabels)
	 */
	public void setDataLabels(DataLabels dataLabels) {
		getPlotOptions().setDataLabels(dataLabels);
	}

	/**
	 * Returns the option of data labels.
	 * 
	 * @see SeriesPlotOptions#getDataLabels()
	 */
	public DataLabels getDataLabels() {
		return getPlotOptions().getDataLabels();
	}

	/**
	 * Returns individual color for the series. By default the color is pulled
	 * from the global <code>colors</code> array.
	 * <p>
	 * Default: null.
	 */
	public Color getColor() {
		return (Color) getAttr(DAttrs.color, null).asValue();
	}

	/**
	 * Sets individual color for the series. By default the color is pulled from
	 * the global <code>colors</code> array.
	 */
	public void setColor(Color color) {
		setAttr(DAttrs.color, color);
	}

	/**
	 * Sets individual color for the series. By default the color is pulled from
	 * the global <code>colors</code> array.
	 */
	public void setColor(String color) {
		setColor(new Color(color));
	}

	/**
	 * Sets individual color for the series. By default the color is pulled from
	 * the global <code>colors</code> array.
	 */
	public void setColor(LinearGradient color) {
		setColor(new Color(color));
	}

	/**
	 * Sets individual color for the series. By default the color is pulled from
	 * the global <code>colors</code> array.
	 */
	public void setColor(RadialGradient color) {
		setColor(new Color(color));
	}
	
	/**
	 * Returns whether color by point option is enabled or not
	 */
	public boolean isColorByPoint() {
		return getAttr(DAttrs.colorByPoint, false).asBoolean();
	}
	
	/**
	 * Sets whether to enable the color by point option or not
	 */
	public void setColorByPoint(boolean colorByPoint) {
		setAttr(DAttrs.colorByPoint, colorByPoint);
	}

	/**
	 * Sets the sequential index of the series in the legend.
	 */
	public void setLegendIndex(int legendIndex) {
		setAttr(DAttrs.legendIndex, legendIndex);
	}

	/**
	 * Returns the sequential index of the series in the legend.
	 */
	public int getLegendIndex() {
		return getAttr(DAttrs.legendIndex, -1).asInt();
	}
	
	/**
	 * Returns the width of the point marker's outline.
	 */
	public Number getLineWidth() {
		return getAttr(DAttrs.lineWidth, null).asNumber();
	}

	/**
	 * Sets the width of the point marker's outline.
	 */
	public void setLineWidth(Number lineWidth) {
		setAttr(DAttrs.lineWidth, lineWidth, null);
	}

	/**
	 * Returns the marker for this series.
	 */
	public Marker getMarker() {
		Marker marker = (Marker) getAttr(DAttrs.marker);
		if (marker == null) {
			marker = new Marker();
			setAttr(DAttrs.marker, marker);
		}
		return marker;
	}

	/**
	 * Sets the marker for this series.
	 */
	public void setMarker(Marker marker) {
		setAttr(DAttrs.marker, marker);
	}

	/**
	 * Sets the name of the series as shown in the legend, tooltip etc.
	 */
	public void setName(String name) {
		setAttr(DAttrs.name, name);
	}
	
	/**
	 * Returns the amount of degrees that are allowed to overshoot
	 * Default: null
	 */
	public Number getOvershoot() {
		return getAttr(DAttrs.overshoot, null).asNumber();
	}
	
	/**
	 * Sets the amount of degrees that are allowed to overshoot
	 */
	public void setOvershoot(Number overshoot) {
		setAttr(DAttrs.overshoot, overshoot);
	}

	/**
	 * Returns the point interval of the series.
	 * Default: null
	 */
	public Number getPointInterval() {
		return getAttr(DAttrs.pointInterval, null).asNumber();
	}
	
	/**
	 * Sets the point interval of the series.
	 */
	public void setPointInterval(Number pointInterval) {
		setAttr(DAttrs.pointInterval, pointInterval);
	}
	
	/**
	 * Returns the padding between each column or bar, in x axis units.
	 * Default: 0.1
	 */
	public Number getPointPadding() {
		return getAttr(DAttrs.pointPadding, 0.1).asNumber();
	}
	
	/**
	 * Sets the padding between each column or bar, in x axis units.
	 */
	public void setPointPadding(Number pointPadding) {
		setAttr(DAttrs.pointPadding, pointPadding, 0.1);
	}
	
	/**
	 * Returns the point placement.
	 * <p>
	 * Possible values: null, "on", "between" or numeric.
	 * </p>
	 * <p>
	 * In a column chart, when pointPlacement is "on", the point will not create
	 * any padding of the X axis. In a polar column chart this means that the
	 * first column points directly north. If the pointPlacement is "between",
	 * the columns will be laid out between ticks. This is useful for example
	 * for visualising an amount between two points in time or in a certain
	 * sector of a polar chart.
	 * </p>
	 * <p>
	 * The point placement can also be numeric, where 0 is on the axis value,
	 * -0.5 is between this value and the previous, and 0.5 is between this
	 * value and the next. Unlike the textual options, numeric point placement
	 * options won't affect axis padding.
	 * </p>
	 * <p>
	 * Defaults to <code>null</code> in cartesian charts, <code>"between"</code>
	 * in polar charts.
	 */
	public Object getPointPlacement() {
		return getAttr(DAttrs.pointPlacement, null).asValue();
	}

	/**
	 * Sets the point placement.
	 * <p>
	 * Possible values: null, "on", "between".
	 * </p>
	 * <p>
	 * In a column chart, when pointPlacement is "on", the point will not create
	 * any padding of the X axis. In a polar column chart this means that the
	 * first column points directly north. If the pointPlacement is "between",
	 * the columns will be laid out between ticks. This is useful for example
	 * for visualising an amount between two points in time or in a certain
	 * sector of a polar chart.
	 * <p>
	 * Defaults to <code>null</code> in cartesian charts, <code>"between"</code>
	 * in polar charts.
	 */
	public void setPointPlacement(String pointPlacement) {
		if (!("on".equals(pointPlacement) || "between".equals(pointPlacement)))
			throw new IllegalArgumentException("Unsupported pointPlacement: ["
					+ pointPlacement + "]");
		setAttr(DAttrs.pointPlacement, pointPlacement);
	}

	/**
	 * Sets the point placement with a number
	 * <p>
	 * The point placement can also be numeric, where 0 is on the axis value,
	 * -0.5 is between this value and the previous, and 0.5 is between this
	 * value and the next. Unlike the textual options, numeric point placement
	 * options won't affect axis padding.
	 * </p>
	 */
	public void setPointPlacement(Number pointPlacement) {
		setAttr(DAttrs.pointPlacement, pointPlacement);
	}

	/**
	 * Returns the name of the series as shown in the legend, tooltip etc.
	 */
	public String getName() {
		return getAttr(DAttrs.name, null).asString();
	}


	/**
	 * Sets the option allows grouping series in a stacked chart. The stack
	 * option can be a string or a number or anything else, as long as the
	 * grouped series' stack options match each other.
	 */
	public void setStack(String stack) {
		setAttr(DAttrs.stack, stack);
	}
	
	/**
	 * Sets the option allows grouping series in a stacked chart. The stack
	 * option can be a string or a number or anything else, as long as the
	 * grouped series' stack options match each other.
	 */
	public void setStack(Number stack) {
		setAttr(DAttrs.stack, stack);
	}

	/**
	 * Returns the option allows grouping series in a stacked chart. The stack
	 * option can be a string or a number or anything else, as long as the
	 * grouped series' stack options match each other.
	 */
	public Object getStack() {
		return getAttr(DAttrs.stack, null).asValue();
	}

	/**
	 * Returns the tooltip options.
	 */
	public Tooltip getTooltip() {
		Tooltip tooltip = (Tooltip) getAttr(DAttrs.tooltip);
		if (tooltip == null) {
			tooltip = new Tooltip();
			setTooltip(tooltip);
		}
		return tooltip;
	}
	
	/**
	 * Sets the tooltip options.
	 */
	public void setTooltip(Tooltip tooltip) {
		setAttr(DAttrs.tooltip, tooltip);
	}
	
	
	/**
	 * The type of series. Can be one of area, areaspline, bar, column, line,
	 * pie, scatter, spline, arearange, areasplinerange and columnrange.
	 */
	public void setType(String type) {
		setAttr(DAttrs.type, type);
	}

	/**
	 * The type of series. Can be one of area, areaspline, bar, column, line,
	 * pie, scatter, spline, arearange, areasplinerange and columnrange.
	 */
	public String getType() {
		return getAttr(DAttrs.type, null).asString();
	}

	/**
	 * When using dual or multiple x axes, this number defines which xAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the xAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public void setXAxis(int xAxis) {
		setAttr(DAttrs.xAxis, xAxis);
	}

	/**
	 * When using dual or multiple x axes, this number defines which xAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the xAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public void setXAxis(String xAxis) {
		setAttr(DAttrs.xAxis, xAxis);
	}

	/**
	 * When using dual or multiple x axes, this number defines which xAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the xAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public Object getXAxis() {
		return getAttr(DAttrs.xAxis, 0).asValue();
	}

	/**
	 * When using dual or multiple y axes, this number defines which yAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the yAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public void setYAxis(int yAxis) {
		setAttr(DAttrs.yAxis, yAxis);
	}

	/**
	 * When using dual or multiple y axes, this number defines which yAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the yAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public void setYAxis(String yAxis) {
		setAttr(DAttrs.yAxis, yAxis);
	}

	/**
	 * When using dual or multiple y axes, this number defines which yAxis the
	 * particular series is connected to. It refers to either the axis id or the
	 * index of the axis in the yAxis array, with 0 being the first. Defaults to
	 * 0.
	 */
	public Object getYAxis() {
		return getAttr(DAttrs.yAxis, 0).asValue();
	}

	/**
	 * Define the visual z index of the series.
	 */
	public void setZIndex(int zIndex) {
		setAttr(DAttrs.zIndex, zIndex);
	}

	/**
	 * Define the visual z index of the series.
	 */
	public int getZIndex() {
		return getAttr(DAttrs.zIndex, -1).asInt();
	}

	/**
	 * Hides the series if visible. If the {@link Chart#isIgnoreHiddenSeries()}
	 * option is true,the chart is redrawn without this series.
	 */
	public void hide() {
		setAttr(Attrs.visible, false);
		fireEvent(EventType.CHANGED, "hide", this, new DeferredCall() {
			public void execute(JSFunction func) {
				func.callFunction("hide");
			}
		});
	}

	/**
	 * Returns whether the series is visible or not.
	 * <p>
	 * Default: true
	 */
	public boolean isVisible() {
		return getAttr(Attrs.visible, true).asBoolean();
	}

	/**
	 * Sets whether the serries is visible.
	 */
	public void setVisible(final boolean visible) {
		if (setAttr(Attrs.visible, visible, true)) {
			fireEvent(EventType.CHANGED, "visible", this, new DeferredCall() {
				public void execute(JSFunction func) {
					func.callFunction("setVisible", visible);
				}
			});
		}
	}

	/**
	 * Shows the series if hidden.
	 */
	public void show() {
		setAttr(Attrs.visible, true);
		fireEvent(EventType.CHANGED, "show", this, new DeferredCall() {
			public void execute(JSFunction func) {
				func.callFunction("show");
			}

		});
	}

	public void remove() {
		fireEvent(EventType.DESTROYED, "series", this, new DeferredCall() {
			public void execute(JSFunction func) {
				func.callFunction("remove");
			}
		});
		clearOptonDataListener();
	}

	public void select() {
		fireEvent(EventType.CHANGED, "select", this, new DeferredCall() {
			public void execute(JSFunction func) {
				func.callFunction("select");
			}

		});
	}

	public void onChange(OptionDataEvent event) {
		final Optionable target = event.getTarget();
		event.setCurrentTarget(this);
		if (target instanceof Point) {
			final int index = getData().indexOf(target);
			if (event.hasJSFunctionCall()) {
				event.addJSFunctionCall(new DeferredCall() {
					public void execute(JSFunction func) {
						func.callArray("data", index);
					}

				});
			}
			if (event.getType() == OptionDataEvent.EventType.DESTROYED) {
				List<Point> list = getData();
				if (list != null)
					list.remove(index);
			}
		} else {
			event.addJSFunctionCall(new DeferredCall() {
				public void execute(JSFunction func) {
					if (target instanceof DataLabels) {
						JSONObject json = new JSONObject();
						json.put("dataLabels", target);
						func.callFunction("update", json);
					} else {
						func.callFunction("update", target);
					}
				}
			});
		}

		fireEvent(event);
	}

	public String toJSONString() {
		if (plotOptions != null) {
			// we don't use merge(options) here
			Map<String, AnyVal<Object>> clone = new LinkedHashMap<String, AnyVal<Object>>(
					options);
			clone.putAll(plotOptions.options);
			return JSONObject.toJSONString(clone);
		}
		return super.toJSONString();
	}
}
